// Copyright 2019 Patric Ljung
// Linkoping University, Media and Information Technology, Visualization Center C
// License: BSD-3-Clause (https://opensource.org/licenses/BSD-3-Clause)


#include "DomeDisplayComponent.h"
#include "Math/UnrealMath.h"
#include "DomeViewerVR.h" // For the log category

//UDomeDisplayComponent::UDomeDisplayComponent(const FObjectInitializer & ObjectInitializer)
//: Super(ObjectInitializer)
//{
//    Radius = 750.0f;
//    TiltAngle = 27.0f;
//    HemisphereDegrees = 170.0f;
//    MeshEdgeDegrees = 3.6f;
//    Material = nullptr;
//
//    BuildDomeMesh();
//
//    bUseComplexAsSimpleCollision = true; // Since it is concave
//
//    CreateMeshSection(0, MeshVerts, MeshTris, MeshNormals, MeshUVs, MeshColors, MeshTangents, true);
//}

UDomeDisplayComponent::UDomeDisplayComponent()
{
    Radius = 750.0f;
    TiltAngle = 27.0f;
    HemisphereDegrees = 170.0f;
    MeshEdgeDegrees = 3.6f;
    Material = nullptr;
    
    DomeMesh = CreateDefaultSubobject<UProceduralMeshComponent>(TEXT("DomeMesh"));
    DomeMesh->SetupAttachment(this);

    //BuildDomeMesh();
    
    //bUseComplexAsSimpleCollision = true; // Since it is concave
    
    //CreateMeshSection(0, MeshVerts, MeshTris, MeshNormals, MeshUVs, MeshColors, MeshTangents, true);
}

UDomeDisplayComponent::~UDomeDisplayComponent()
{
    Material = nullptr;
}

#define PI_2 (2.0f * PI)

void UDomeDisplayComponent::BuildDomeMesh()
{
    int numLatSections = 1 + int(HemisphereDegrees / MeshEdgeDegrees / 2.0f);
    int numLonSections = 1 + int(360.0f / MeshEdgeDegrees);
    
    // DEBUG: Just make it simpler for debugging
    //numLatSections = 2;
    
    UE_LOG(DomeViewer, Display, TEXT("BuildDomeMesh: LatSections %d, LonSections: %d"), numLatSections, numLonSections);

    const FVector forward = FVector(1.0f, 0.0f, 0.0f);
    const FVector right = FVector(0.0f, 1.0f, 0.0f);
    const FVector up = FVector(0.0f, 0.0f, 1.0f);
    
    int numVerts = 1 + numLatSections * numLonSections;
    MeshVerts.Reserve(numVerts);
    MeshUVs.Reserve(numVerts);
    MeshTris.Reserve(3*numVerts); // TODO: Resolve real number of triangles
    
    const float deltaLatRads = (PI * HemisphereDegrees / 360.0f) / float(numLatSections - 1);
    const float deltaLonRads = PI_2 / float(numLonSections);
    
    // Compute the vertices, starting at top
    MeshVerts.Add(up * Radius);
    MeshNormals.Add(-up);
    //MeshTangents.Add(right);

    for (int latIdx = 1; latIdx < numLatSections; ++latIdx)
    {
        float latAngle = latIdx * deltaLatRads;
        float latSin = sin(latAngle);
        float latCos = cos(latAngle);
        FVector upLoc = up * latCos;
        
        for (int lonIdx = 0; lonIdx < numLonSections; ++lonIdx)
        {
            float lonSin = sin(lonIdx * PI_2 / numLonSections);
            float lonCos = cos(lonIdx * PI_2 / numLonSections);
            FVector unitDir(forward * (latSin * lonCos) +
                            right * (latSin * lonSin) + upLoc);
            MeshVerts.Add(unitDir * Radius);
            MeshNormals.Add(-unitDir);
        }
    }
    
    // Compute mesh UVs, starting at the center
    const FVector2D uvCenter(0.5f, 0.5f);
    MeshUVs.Add(uvCenter);
    
    for (int latIdx = 1; latIdx < numLatSections; ++latIdx)
    {
        float uvRadius = 0.5f * float(latIdx) / float(numLatSections - 1);

        for (int lonIdx = 0; lonIdx < numLonSections; ++lonIdx)
        {
            float lonAngle = lonIdx * deltaLonRads;
            MeshUVs.Add(uvCenter + uvRadius * FVector2D(sin(lonAngle), cos(lonAngle)));
        }
    }

    // Compute the triangle indices
    // ... first the cap
    for (int lonIdx = 0; lonIdx < numLonSections; ++lonIdx)
    {
        // Counter clock-wise, triangle fan
        MeshTris.Add(1 + lonIdx);
        MeshTris.Add(1 + ((lonIdx + 1) % numLonSections));
        MeshTris.Add(0);
    }
    // ... then the rest
    for (int latIdx = 1; latIdx < numLatSections - 1; ++latIdx)
    {
        int32 topBase = 1 + (latIdx - 1) * numLonSections;
        int32 bottomBase = 1 + latIdx * numLonSections;

        for (int lonIdx = 0; lonIdx < numLonSections; ++lonIdx)
        {
            // Make quads, two triangles
            // Counter clock-wise
            MeshTris.Add(bottomBase + lonIdx);
            MeshTris.Add(bottomBase + (lonIdx + 1) % numLonSections);
            MeshTris.Add(topBase + lonIdx);

            MeshTris.Add(bottomBase + (lonIdx + 1) % numLonSections);
            MeshTris.Add(topBase + (lonIdx + 1) % numLonSections);
            MeshTris.Add(topBase + lonIdx);
        }
    }
}

bool UDomeDisplayComponent::UpdateDomeMesh()
{
    BuildDomeMesh();

    DomeMesh->CreateMeshSection(0, MeshVerts, MeshTris, MeshNormals, MeshUVs, MeshColors, MeshTangents, false);

    if (Material != nullptr)
    {
        DomeMesh->SetMaterial(0, Material);
    }

    DomeMesh->SetRelativeRotation(FRotator(-TiltAngle, 0.0f, 0.0f));

    return true;
}
